/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "jomar3D.h"
#if (USEGLEW)
#  include <GL/glew.h>
#endif
#include <GL/glut.h>

#include "display.h"
#include "draw.h"
#include "drawmenu.h"
#include "drawtext.h"
#include "drawworld.h"
#include "interface.h"
#include "readpng.h"
#include "scores.h"
#include "sound.h"
#include "utils.h"

/* Position of the light either normal or when drawing the menu */
static GLfloat l_lightPosition[] = { 0.0, 9000.0, 0.0, 1.0 };
static GLfloat l_lightPositionmenu[] = { -100.0, 1000.0, -100.0, 1.0 };

static GLuint l_texsid[TEXID_N_ID];  /* IDs of all textures */

static GLfloat l_viewrot;   /* rotation of the whole view */
static Coord3f l_angle;     /* angle of the whole view in degrees */
static Coord3f l_pose;      /* position of the eyes to the center of the lobster */
static Coord2f l_poseye;    /* position in x and y of the eye of the player */
static float l_gonev;       /* "gone" a little changed that is used for the view  */

#define LEGLONG 26.5        /* length of each leg */
static float l_legp[4] = { 0., 0., 0., 0. };      /* position of the legs from the begin of the move */
static float l_legb[4] = { -15., -5., 5., 25. };  /* base angle for legs */

static GLUquadric *l_quadObj;     /* the only quadric used */

#define SELBUFFLEN 512                /* the length of the selection buffer */
static GLuint l_selbuff[SELBUFFLEN];  /* selection buffer */

static int l_measuretime = 0;     /* time of teh last measure of fps */
static int l_frames = 0;          /* frames counter, incremented at each frame */
static int l_fps;                 /* current fps */
static float l_average;           /* average of fps */

static char *l_music    = NULL;   /* name of the current music (to be shown) */

#define INTRODURATION 1000        /* duration of the introduction */
static int l_introstarttime = 0;  /* time at which the introduction started */
static int l_introelapsed = 0;    /* time elapsed since beginning of introduction */

static Bomb l_bomb1;              /* storage of all variables defining a bomb */

static int l_vertfact = 1;        /* factor to multiply the vertex number to reach the desired quality */
static int l_quality = 1;         /* current quality set */


/* Quit cleanly the drawing modules */
void
draw_quit (void)
{
  /* Quit submodules */
  drawmenu_quit ();
  drawworld_clearWells ();

  free (l_music);
  gluDeleteQuadric (l_quadObj);
}

/* Generate a full redraw */
void
draw_forceRedraw (void)
{
  if (glutGetWindow()) /* So it could be called if the display isn't ready */
    glutPostRedisplay ();
}


/* Give the current quality */
int
draw_getQuality (void)
{
  return l_quality;
}

/* Give the used GLU quadric's pointer */
GLUquadric*
draw_getQuadric (void)
{
  return l_quadObj;
}

/* Give the OpenGL texture ID of the given texture  */
int
draw_getTexID (int name)
{
  return l_texsid[name];
}

/* Give the multiplicator of vertices set by the quality */
int
draw_getVertMult (void)
{
  return l_vertfact;
}

/* Say if the intro if off */
int
draw_postIntro (void)
{
  return (l_introelapsed < INTRODURATION) ? 0 : 1;
}

/*
 *  Function to call to manage the quality settings
 * 
 *    action: 0 = use the given value without thinking too much
 *            1 = modify the current one from 'value' steps (1 OR -1)
 */
void
draw_applyQuality (int action, int value)
{
  if (action)
  {
    l_quality += value;
    if (l_quality < Q_FINE)
      l_quality = Q_HORRIBLE;
    if (l_quality > Q_HORRIBLE)
      l_quality = Q_FINE;
  }
  else
    l_quality = value;

  if (l_quality < Q_FINE || l_quality > Q_HORRIBLE)
  {
    fprintf (stderr, "Unexpected quality set, assuming 'normal'\n");
    l_quality = Q_NORMAL;
  }

  drawworld_setWellsDividers ((2 - l_quality + 1) * (2 - l_quality + 1));
  
  /* sets quality vars if choosen quality is @p quality
   * WARNING: this macto is NOT safely usable outside of its strict context */
  #define SETQUALITYIF(quality, vfact, text) \
    if (l_quality == quality) { \
      l_vertfact = vfact; \
      core_setQualityText (text); \
    }
  
  SETQUALITYIF (Q_HORRIBLE, 1, "horrible")
  SETQUALITYIF (Q_NORMAL,   2, "normal")
  SETQUALITYIF (Q_FINE,     4, "fine")
  
  #undef SETQUALITYIF

  drawworld_clearWells ();
  drawworld_makeWellsData ();
  drawworld_activateArray ();

  drawworld_makeAll ();
}

/* calculate the frames drawn since last call */
void
draw_calculateFPS (int flag /* flag meaning "reset" */)
{
  int now     = glutGet (GLUT_ELAPSED_TIME);
  int elapsed = now - l_measuretime;

  if (flag)
  {
    l_average = 0;
    l_fps = 0;
  }
  else
  {
    l_fps = (int) (l_frames * 1000. / elapsed + 0.5);
    if (l_fps != 0)
      l_average = (l_average * 20. + (float) (l_fps)) / 21.;
    l_measuretime = now;
    l_frames = 0;
  }
}


/* Do what should only be done once per load of the game */
void
draw_init (void)
{
  int i;
#if USEGLEW && defined (GL_ARB_vertex_buffer_object)
  GLenum err;
  const unsigned char *exts;
#endif /* USEGLEW && defined (GL_ARB_vertex_buffer_object) */

/* initialize the quadric */
  l_quadObj = gluNewQuadric ();

/* initialize the legs' position */
  for (i = 0; i <= 3; i++)
    core_getLobster ()->legangle[i] = l_legb[i];

/* Initialize the name if the music to be shown */
  l_music = mystrdup ("X");

/* Initialize submodules */
  drawmenu_initCyls ();

/* Initialze the extension management */
#if USEGLEW && defined (GL_ARB_vertex_buffer_object)
  if ((err = glewInit ()) != GLEW_OK)
  {
    /* Problem: glewInit failed, something is seriously wrong. */
    fprintf (stderr, "Error: %s\n", glewGetErrorString (err));
  }
  fprintf (stdout, "Status: Using GLEW %s\n", glewGetString (GLEW_VERSION));
  
  /* get the supported extensions list */
  exts = glGetString (GL_EXTENSIONS);

  /* look for the desired extension */
  if (strstr ((const char *) exts, "GL_ARB_vertex_buffer_object") == NULL)
  {
    /* extension not supported... :-( */
    fprintf (stderr,
             "Extension '%s' not supported, but continuing happily without...\n",
             "GL_ARB_vertex_buffer_object");
  }
  else
  {
    /* OK, extension supported */
    fprintf (stdout, "Extension '%s' is supported, using it...\n",
             "GL_ARB_vertex_buffer_object");
  }
#else /* !USEGLEW || !defined (GL_ARB_vertex_buffer_object) */
  fprintf (stdout, "Extension '%s' is not built in...\n",
           "GL_ARB_vertex_buffer_object");
#endif /* glew test */
}

/* say if the given x, y (in squares) surrounded by border (in gl units) is seen */
int
draw_isVisible (int x, int y, float border)
{
  float dangle; /* angle to use */
  float xw;
  float yw;
  
  dangle = -(l_angle.x - core_getLobster ()->angle) * degrad;
  xw = x*T_C + T_C/2. + l_poseye.x;
  yw = y*T_C + T_C/2. - l_poseye.y;
  
  /*if (y > 15)
    return TRUE;*/
/* try the four edges of the area to test
 * just see if after rotation and translation the point it is in positive y */

  if ((xw + border) * sin(dangle) + (yw + border) * cos(dangle) > 0.)
    return TRUE;
  if ((xw + border) * sin(dangle) + (yw - border) * cos(dangle) > 0.)
    return TRUE;
  if ((xw - border) * sin(dangle) + (yw - border) * cos(dangle) > 0.)
    return TRUE;
  if ((xw - border) * sin(dangle) + (yw + border) * cos(dangle) > 0.)
    return TRUE;
  
  return FALSE;
}


/* Turn the view to see the exploding bomb. Must be called if the player lost. */
static void
draw_turnToBomb (void)
{
  float angle2bomb;
  float lobangle = core_getLobster ()->angle;
  
  angle2bomb = acos ((l_poseye.x + l_bomb1.pos.x*T_C) /
                     sqrt(pow(l_poseye.x + l_bomb1.pos.x*T_C, 2) + pow(l_poseye.y - l_bomb1.pos.y*T_C, 2)))
                     *raddeg;
  if (l_poseye.y - l_bomb1.pos.y*T_C < 0.)
    angle2bomb = 360 - angle2bomb;
  angle2bomb += 90.;

  /* Make sure the angle of the view wouldn't make us sick by turning
   * too much : more than a half turn.
   * So make it fit to desired angle ± 180 degrees */

  retry1:
  if (l_angle.x < lobangle - angle2bomb - 180.)
  {
    l_angle.x += 360.;
    goto retry1;
  }
  retry2:
  if (l_angle.x >= lobangle - angle2bomb + 180.)
  {
    l_angle.x -= 360.;
    goto retry2;
  }

  l_angle.x += ((lobangle - angle2bomb) - l_angle.x) / (float) l_fps;
  l_angle.y -= l_angle.y / (float) l_fps;
}


/* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX (end drawing initialization & exec) */

/* draw the logo on top -- must be called the good matrix loaded */
static void
draw_drawIntro (void)
{
  float pos;

  if (l_introstarttime == 0)
    l_introstarttime = glutGet (GLUT_ELAPSED_TIME);

  l_introelapsed = glutGet (GLUT_ELAPSED_TIME) - l_introstarttime;
  if (l_introelapsed > INTRODURATION)
    return;

  glColor4f (0.0, 0.0, 0.0, 1.0);
  glEnable (GL_TEXTURE_2D);
  glBindTexture (GL_TEXTURE_2D, l_texsid[TEXID_LOGO]);
  if (l_introelapsed < INTRODURATION / 2.)
  {
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    pos = 1500. * (float) (l_introelapsed) / ((float) (INTRODURATION) / 2.);
    glBegin (GL_QUADS);
    glTexCoord2f (0.0, 0.0);
    glVertex2f (0.0, 0.0);
    glTexCoord2f (1.0, 0.0);
    glVertex2f (pos, 0.0);
    glTexCoord2f (1.0, 1.0);
    glVertex2f (pos, pos);
    glTexCoord2f (0.0, 1.0);
    glVertex2f (0.0, pos);
    glEnd ();
  }
  else
  {
    pos =
      1500. * ((float) (l_introelapsed) -
               (float) (INTRODURATION) / 2.) / ((float) (INTRODURATION) /
                                                2.);
    glBegin (GL_QUADS);
    glTexCoord2f (0.0, 0.0);
    glVertex2f (pos, pos);
    glTexCoord2f (1.0, 0.0);
    glVertex2f (1500, pos);
    glTexCoord2f (1.0, 1.0);
    glVertex2f (1500, 1500);
    glTexCoord2f (0.0, 1.0);
    glVertex2f (pos, 1500);
    glEnd ();
  }

  glDisable (GL_TEXTURE_2D);
  glutPostRedisplay ();
}

/* Draw the minimap in the bottom left part of the screen */
static void
draw_drawMinimap (void)
{
  float dangle        = (int)-(l_angle.x - core_getLobster ()->angle)%360;
  Coord2i *lobsterpos = &(core_getLobster ()->pos);
  
  /*printf ("%f\n", dangle);*/
  glPushMatrix ();
  display_setDefault2D ();

  if ((dangle >= 0 && dangle < 90) ||
      (dangle >= -360 && dangle < -270))
    glTranslatef (T_Y * SM_PITCH * sin(dangle * degrad),  40., 0.);
  if ((dangle >= 90 && dangle < 180) ||
      (dangle >= -270 && dangle < -180))
    glTranslatef (sin(atan((float) T_Y / (float) T_X) + (dangle - 90) * degrad) *
      sqrt (T_Y*SM_PITCH * T_Y*SM_PITCH + T_X*SM_PITCH * T_X*SM_PITCH),
      40. - T_Y * SM_PITCH * cos(dangle * degrad), 0.);
  if ((dangle >= 180 && dangle < 270) ||
      (dangle >= -180 && dangle < -90))
    glTranslatef (-T_X * SM_PITCH * cos(dangle * degrad),
      40. +
      cos (atan((float) T_Y / (float) T_X) + (dangle - 270.) * degrad) *
      sqrt (T_Y*SM_PITCH * T_Y*SM_PITCH + T_X*SM_PITCH * T_X*SM_PITCH),
      0.);
  if ((dangle >= 270 && dangle < 360) ||
      (dangle >= -90 && dangle < 0))
    glTranslatef (0., -T_X * SM_PITCH * sin(dangle * degrad) + 40., 0.);

  glRotatef (dangle, 0., 0., 1.);
  glCallList (SMALLMAP);
  glColor4f (0.0, 0.0, 1.0, 1.0);
  glRecti (lobsterpos->x * SM_PITCH, lobsterpos->y * SM_PITCH,
           (lobsterpos->x + 1) * SM_PITCH, (lobsterpos->y + 1) * SM_PITCH);
  
#if (DEBUG)
  /* draw martians in debugging mode */
  {
    int i;
    
    for (i = 0; i < NMARTIANS; i++)
    {
      Martian *mart = core_getMartian (i);
      if (mart && mart->speed != 0)
      {
        glColor4f (0.0, 1.0, 0.0, 1.0);
        glRecti (mart->coords.x * SM_PITCH, mart->coords.y * SM_PITCH,
                 (mart->coords.x + 1) * SM_PITCH, (mart->coords.y + 1) * SM_PITCH);
      }
    }
  }
#endif
  
  if (l_bomb1.throwtime)
  {
    glColor4f (1.0, 0.0, 0.0, 1.0);
    glRecti (l_bomb1.pos.x * SM_PITCH, l_bomb1.pos.y * SM_PITCH,
             (l_bomb1.pos.x + 1) * SM_PITCH, (l_bomb1.pos.y + 1) * SM_PITCH);
  }
  glPopMatrix ();
}

/* draw what needed by the score interface */
static void
draw_drawScores (void)
{
  int i;
  char *buf;
  Player *playerdrawn;

  glColor4f (0.0, 0.0, 1.0, 1.0);
  drawtext_drawString (750., 200., 1.0, 1.0, TEXT_CENTER, 1400., FALSE, TRUE,
                       scores_getCurrentName ());

  /* If we haven't validated in scores then we must simply draw
   * the screen of asking a name to use */ 
  if (!scores_validateName ())
  {
    drawtext_drawString (750., 1200., 0.5, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                         core_getTextPointer(29));
    if (scores_getName(0) != NULL)
      drawtext_drawString (750., 1100., 0.5, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                           core_getTextPointer(30));
    for (i = 0; i < MAXNAMES && /*strlen (scores_getName(i)) != 0*/ scores_getName(i) != NULL; i++)
    {
      playerdrawn = scores_getPlayer (i);
      if (!playerdrawn)
        continue;
      
      buf = malloc (strlen (playerdrawn->name) + strlen (core_getTextPointer(32)) + 1 + intlen (playerdrawn->games, 10) +1);
      if (buf)
      {
        sprintf (buf, "%s%s %d", playerdrawn->name, core_getTextPointer(32), playerdrawn->games);
        drawtext_drawString (750., 1000. - (700. / MAXNAMES) * i, 0.3, 0.3,
                             TEXT_CENTER, 1400., FALSE, TRUE,
                             buf);
        
        free (buf);
      }
      else
        perror ("failed to allocate memory");
    }
  }
}

/* print all the texts needed by the various states */
static void
draw_drawGame (void)
{
  char *buf;
  char *txt;
  int timeelapsed = core_getGameDuration ();

  drawtext_begin ();

  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();
  gluOrtho2D (0, 1500, 0, 1500);

  if (core_checkState (JOMAR3D_END))
  {
    if (core_checkState (JOMAR3D_SCORE))
    {
      draw_drawScores ();
    }
    else
    {
      glColor4f (1.0, 0.0, 0.0, 0.8);
      glLineWidth (5);
      drawtext_drawString (750., 750., 1., 1., TEXT_CENTER, 1400., FALSE, TRUE,
                           core_getTextPointer(2));
      glLineWidth (2);

      txt = time2s (timeelapsed / 1000.);
      
      /* allocation for "%s %.3f %s" */
      buf = malloc (strlen (core_getTextPointer(3)) + 1 +
                    strlen (txt) + 1 +
                    strlen (core_getTextPointer(4)) +1);
      if (buf)
      {
        sprintf (buf, "%s %s %s", core_getTextPointer(3), txt,
                 core_getTextPointer(4));
        drawtext_drawString (750., 600., 0.4, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                             buf);
      }
      else
        perror ("failed to allocate memory");
      free (buf);
      
#if (USESOUND)
      sound_mix (WON);
#endif
    }
  }
  glLineWidth (2);
  glColor4f (1.0, 1.0, 1.0, 0.5);
  if (!core_checkState (JOMAR3D_INMENU))
  {
    if (core_checkState (JOMAR3D_LAST))
    {
      buf = malloc (strlen (core_getTextPointer(1)) +1);
      if (buf)
        sprintf (buf, "%s", core_getTextPointer(1));
      else
        perror ("failed to allocate memory");
    }
    else
    {
      /* allocation for "  %s %d  " */
      buf = malloc (2 + strlen (core_getTextPointer(0)) + 1 + intlen (core_getMartiansLeft (), 10) + 2 +1);
      if (buf)
        sprintf (buf, "  %s %d  ", core_getTextPointer(0), core_getMartiansLeft ());
      else
        perror ("failed to allocate memory");
    }
    drawtext_drawString (1500. / 2., 10, 1.15, 0.3, TEXT_CENTER, 1500., FALSE, TRUE,
                         buf);
    free (buf);
  }
  if (!core_checkState (JOMAR3D_INMENU))
  {
    /* allocation for "%s %d-%d" */
    buf = malloc (strlen (core_getTextPointer(5)) + 1 + intlen (l_fps, 10) + 1 +
                  intlen ((int) (l_average + 0.5), 10) +1);
    if (buf)
    {
      sprintf (buf, "%s %d-%d", core_getTextPointer(5), l_fps,
               (int) (l_average + 0.5));
      drawtext_drawString (1500, 57, 0.2, 0.2, TEXT_RIGHT, 1500., FALSE, TRUE,
                           buf);
    }
    else
      perror ("failed to allocate memory");
    free (buf);
  }

  glLineWidth (1.8);
  glColor4f (0.0, 0.0, 0.0, 0.5);
  if (core_checkState (JOMAR3D_INMENU))
  {
    
    drawtext_drawString (0., 1422., 0.5, 0.2, TEXT_LEFT, 1500., FALSE, TRUE,
                         "www.");
    drawtext_drawString (750., 1460., 1.15, 0.3, TEXT_CENTER, 1300., FALSE, TRUE,
                         "Studios.Chalmion");
    drawtext_drawString (1500., 1422., 0.5, 0.2, TEXT_RIGHT, 1500., FALSE, TRUE,
                         ".tuxfamily.org");
  }
  else
  {
    txt = time2s (timeelapsed / 1000);
    drawtext_drawString (1491, 1461, 0.5, 0.3, TEXT_RIGHT, 1500., FALSE, TRUE,
                         txt);
    free (txt);
    
    buf = malloc (uintlen (core_getMartianSteps (), 10) +1);
    if (buf)
    {
      sprintf (buf, "%i", core_getMartianSteps ());
      drawtext_drawString (1491, 1410, 0.5, 0.3, TEXT_RIGHT, 1500., FALSE, TRUE,
                           buf);
      free (buf);
    }
    else
      perror ("failed to allocate memory");
    
#if (USESOUND)
    if (core_checkState(JOMAR3D_PLAYMUSIC))
      drawtext_drawString (0, 1475, 0.2, 0.2, TEXT_LEFT, 1500., FALSE, TRUE,
                           l_music);
#endif
  }

  if (!core_checkState (JOMAR3D_INMENU))
    draw_drawMinimap ();

  if (l_introelapsed <= INTRODURATION)
    draw_drawIntro ();

  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();

  drawtext_end ();

  /*glPopAttrib ();*/
  glMatrixMode (GL_MODELVIEW);
}

/* Call this to set up the matrix to select */
GLuint *
draw_pickingRender (int x, int y)
{

  glRenderMode (GL_SELECT);
  glInitNames ();
  glPushName (-1);

  /* setup a picking matrix and render into selection buffer */
  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();

  display_setPickport (x, y);

  glMatrixMode (GL_MODELVIEW);
  glCallList (MENU);

  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();
  glMatrixMode (GL_MODELVIEW);
  glPopMatrix ();

  return l_selbuff;
}

/* Called to init the bomb drawing */
void
draw_startBombDraw (int posx, int posy, int duration)
{
  l_bomb1.pos.x = posx;
  l_bomb1.pos.y = posy;
  l_bomb1.angle.x = MYRAND * 30.;
  l_bomb1.angle.z = MYRAND * 30.;
  l_bomb1.throwtime = glutGet(GLUT_ELAPSED_TIME);
  l_bomb1.duration  = duration;
}

/* Manage to do all what needed by the bomb */
static void
draw_drawBomb (void)
{
  int elapsed = glutGet(GLUT_ELAPSED_TIME) - l_bomb1.throwtime;

/* draw only if bomb seen or if you lost */
  if (draw_isVisible (l_bomb1.pos.x, l_bomb1.pos.y, T_C) || core_getLost ())
  {
    glDisable (GL_LIGHTING);
    if (elapsed < l_bomb1.duration)
      drawworld_drawBurningBomb (&l_bomb1);
    else
    {
      if (core_getLost ())
        draw_turnToBomb ();
      drawworld_drawExplodingBomb (&l_bomb1);
    }
    glEnable (GL_LIGHTING);
    draw_forceRedraw (); 
  }

  if (elapsed >= l_bomb1.duration + EXPLODE || core_checkState (JOMAR3D_INMENU))
    l_bomb1.throwtime = 0.;

}

/* (re)draw all the window content */
void
draw_redraw (void)
{
  const Martian *mart;
  const Lobster *lobster = core_getLobster ();
  int number;
  /* true angle of the view in GL coordinates */
  float anglexv = l_angle.x - lobster->angle;
  /* alterate a little the angle of the lobster in function of
   * the delta inducted from the current move */ 
  float anglehom2 = lobster->angle - sin((lobster->anglebase - lobster->angle) * 2. * degrad)*15.;

/* don't clear the screen, it's very too long */
  /*
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glClear(GL_DEPTH_BUFFER_BIT);
  */

/* Set the view */
  glPushMatrix ();
  core_computeLobsterPosition ();

  l_poseye.x = -(lobster->pos.x) * T_C +
               l_pose.x * cos(anglehom2*degrad) - (l_pose.y+l_gonev) * sin(anglehom2*degrad);
  l_poseye.y = lobster->pos.y * T_C +
               l_pose.x * sin(anglehom2*degrad) + (l_pose.y+l_gonev) * cos(anglehom2*degrad);

  gluLookAt (l_poseye.x, 50.0+l_pose.z, l_poseye.y,  /* eye   */
             l_poseye.x + 200. * sin(anglexv * degrad), 20.0 + 200.0 * tan(l_angle.y * degrad), l_poseye.y + 200. * cos(anglexv * degrad),
             -sin(l_viewrot) * cos(anglexv * degrad), cos (l_viewrot), sin(l_viewrot) * sin(anglexv * degrad));

/* to have a view from over the map */
/*
  glPopMatrix ();
  glPushMatrix ();
  gluLookAt ( -T_X*T_C/2., 1800., T_Y*T_C/2.,
              -T_X*T_C/2., 0000., T_Y*T_C/2.,
              0., 0., 1.);
  glDisable(GL_CULL_FACE); */


  /* Because "the [light] position is transformed by the
   *          modelview matrix when glLight is called
   *          (just as if it were a point), and it is
   *          stored in eye coordinates",
   * I have to call glLight each time the view changed, what happen
   * almost at each draw_redraw. */
  glLightfv (GL_LIGHT0, GL_POSITION, l_lightPosition);


/*  draw the world at the beginning to permit nice shadows on the floor */

  glDisable (GL_LIGHTING);
  glDepthFunc (GL_ALWAYS);

  glCallList (TOPBOT);

  glDepthMask (GL_FALSE);

/* Draw that sun-like thing if we can see it or if we are crazy with the mouse wheel
 * because then I don't know when it should be shown... */

  if (l_angle.y > 45. || l_viewrot != 0.)
  {
    glPushMatrix ();
      glTranslatef (-(lobster->pos.x) * T_C, 3000.0, lobster->pos.y * T_C);
      glRotatef (-(lobster->angle), 0.0, 1.0, 0.0);
      glTranslatef (0., 0., lobster->gone);
      glRotatef (90, 1.0, 0.0, 0.0);
      glColor4f (.7, .7, .8, 1.);
      gluDisk (l_quadObj, 0, 250., 20, 1);
      glColor4f (1., 1., 0., 1.);
      gluDisk (l_quadObj, 0, 150., 15, 1);
    glPopMatrix ();
 }

/* Draw all what flie over the head */

  glEnable (GL_TEXTURE_2D);
  glCallList (CLOUDS);
  glDepthMask (GL_TRUE);
  glCallList (WALLMAX);
  glDisable (GL_TEXTURE_2D);

  glDepthFunc (GL_LESS);
  glEnable (GL_LIGHTING);


/* Draw the lobster and its shadow */
  glPushMatrix ();
    glTranslatef (-(lobster->pos.x) * T_C, 0.0, lobster->pos.y * T_C);
    glRotatef (-(lobster->angle), 0.0, 1.0, 0.0);
    glTranslatef (0., .0, lobster->gone);

    /*    shadow    */
    glPushMatrix ();

      glEnable (GL_BLEND);
      glDisable (GL_LIGHTING);
      glColor4f (0.0, 0.0, 0.0, 0.3);

      glDepthRange (0.999, 0.9990001);
      glDepthFunc (GL_GREATER);
      /* floorshadow is good but the scaling is enough here and simplier */
      /*glMultMatrixf ((GLfloat *) floorShadow);*/
      glScalef(1., 0.001, 1.);
      drawworld_drawLobster ();
      glDepthFunc (GL_LESS);
      glDepthRange (0., 1.);

      glEnable (GL_LIGHTING);
      glDisable (GL_BLEND);
    glPopMatrix ();

    /* Body */
    glTranslatef (0., LEGLONG + lobster->phase / 10., 0.);
    glShadeModel (GL_SMOOTH);
    drawworld_drawLobster ();
    glShadeModel (GL_FLAT);
  glPopMatrix ();

/* Activate the Martians (look around or move) */
  core_activeMartians ();

/* Draw the Martians  */

  for (number = 0; number < NMARTIANS; number++)
  {
    mart = core_getMartian (number);
    if (mart->speed == 0.) /* Martian was eaten */
      continue;
    if (!draw_isVisible (mart->coords.x, mart->coords.y, T_C/2.))
      continue;   /* Martian is here in the back of the player */
    glPushMatrix ();
    glTranslatef (-(mart->coords.x) * T_C - mart->coordsgl.x, 20.0,
                   mart->coords.y * T_C + mart->coordsgl.y);
    glRotatef (-(mart->angle), 0., 1., 0.);

    glShadeModel (GL_SMOOTH);
    if (mart->state)
      drawworld_drawMartian (ON);
    else
      drawworld_drawMartian (OFF);
    glShadeModel (GL_FLAT);

    glTranslatef (0., -18.0, 0.0);

    /* shadow  */ 
    glEnable (GL_BLEND);
    glDisable (GL_LIGHTING);
    glColor4f (0.0, 0.0, 0.0, 0.3);

    /* Project the shadow. */
    /*glMultMatrixf ((GLfloat *) floorShadow);*/
    glScalef(1., 0.001, 1.);

    if (mart->state)
      drawworld_drawMartian (ON | SHADOW);
    else
      drawworld_drawMartian (OFF | SHADOW);

    glDisable (GL_BLEND);
    glEnable (GL_LIGHTING);
    glPopMatrix ();
  }

  /* Draw the bomb if any */
  if (l_bomb1.throwtime)
    draw_drawBomb();

/*  draw the wells at the end so we can see if Martians are behind the wells (ONLY IF ALPHA) */
  glDisable (GL_LIGHTING);
  /* initialize the color otherwise some wells are sometimes not drawn (!) */
  glColor4f (1.0, 1.0, 1.0, 1.0);

  glEnable (GL_TEXTURE_2D);
  glEnable (GL_ALPHA_TEST);
  glAlphaFunc (GL_GREATER, 0.5);
  glDisable (GL_CULL_FACE);

  drawworld_wells ();

  glDisable (GL_ALPHA_TEST);
  glDisable (GL_TEXTURE_2D);
  glEnable (GL_CULL_FACE);

  glPopMatrix ();
  draw_drawGame ();
  glEnable (GL_LIGHTING);


/* draw the menu after all so it does not change in function of what is behind
 * only if we are in a menu and the intro is finished  */
  if (core_checkState (JOMAR3D_INMENU) && l_introelapsed > INTRODURATION)
  {
    glLightfv (GL_LIGHT0, GL_POSITION, l_lightPositionmenu);

    glDepthFunc (GL_ALWAYS);
    glDepthMask (GL_FALSE);
    glCallList (MENU);
    if (core_checkState (JOMAR3D_MENU_RULES) || core_checkState (JOMAR3D_MENU_FINDS) ||
        core_checkState (JOMAR3D_MENU_DISPSCORE))
      drawmenu_draw ();
    glDepthMask (GL_TRUE);
    glDepthFunc (GL_LESS);
  }

/* Finish and count this frame */
  glutSwapBuffers ();
  l_frames++;

#if (DEBUG)
  glutReportErrors ();
#endif
}

/* Set the array of characters to draw the music's name */
void
draw_setMusicName (const char *text)
{  
  int musiclen;

  /* get length of the music filename without extension */
  musiclen = strrchr (text, '.') - text;
  /* if no extension was found, use the full name */
  if (musiclen < 0) /* C99 defines that a NULL pointer expands to 0 (see 6.3.2.3
                       and 7.17), then this is right */
    musiclen = strlen (text);

  free (l_music);
  l_music = malloc (musiclen + 1);
  if (l_music != NULL)
  {
    strncpy (l_music, text, musiclen);
    l_music[musiclen] = '\0';
  }
}


/* add the directory prefix before calling the function to load textures */

static GLuint
draw_loadMipmaps (const char *imageFile)
{
  char*   tmp;
  GLuint  idgen = 0;

  /* ensure the skin exists */
  if (core_getSkinName () == NULL)
    return 0;

  tmp = malloc (strlen (SKINDIR "/") + strlen (core_getSkinName ()) +
                5 + strlen (imageFile) + 4 + 1);
  if (tmp)
  {
    sprintf(tmp, "%s%s%s%s%s", SKINDIR "/", core_getSkinName (), ".jsk/",
            imageFile, ".png");

    idgen = readpng_load_mipmaps_gl (tmp);

    free(tmp);
  }
  else
    perror ("failed to allocate memory");

  return idgen;
}

/* read the textures */

void
draw_loadTextures (int first /* says if it is the first time we load all*/)
{
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  if (!first)
    glDeleteTextures (TEXID_N_ID, l_texsid);
  /*glGenTextures (TEXID_N_ID, texsid);*/

  /* load mipmaps */
  #define DLOADMIPMAP(arrid, fname) \
    (l_texsid[arrid] = draw_loadMipmaps (fname))

  DLOADMIPMAP (TEXID_GROUND, "ground");

  DLOADMIPMAP (TEXID_EXT, "ext");

  /* simplfied to load wall mipmaps */
  #define DLOADMIPMAP_WALL(walln) \
    (DLOADMIPMAP (TEXID_WALL##walln, "wall"#walln))

  DLOADMIPMAP_WALL (1);
  DLOADMIPMAP_WALL (2);
  DLOADMIPMAP_WALL (3);
  DLOADMIPMAP_WALL (4);
  DLOADMIPMAP_WALL (5);
  DLOADMIPMAP_WALL (6);
  DLOADMIPMAP_WALL (7);
  DLOADMIPMAP_WALL (8);
  DLOADMIPMAP_WALL (9);
  DLOADMIPMAP_WALL (10);

  #undef DLOADMIPMAP_WALL

  DLOADMIPMAP (TEXID_CLOUD, "cloud");
  DLOADMIPMAP (TEXID_LOGO, "logo");

  DLOADMIPMAP (TEXID_MENU, "menu");

  #undef DLOADMIPMAP
}

/* Set all attributes that need to be to start the game */
void
draw_setGameAttributes (void)
{
  GLfloat black[] = { 0., 0., 0., 0. };
  GLfloat texColor[] = { 1.0, 1.0, 1.0, 1.0 };

  /*glClearColor(fogColor[0], fogColor[1], fogColor[2], fogColor[3]);*/
  /*glutIgnoreKeyRepeat(1);*/ /* seems that it losts some keys!*/
  glutDisplayFunc (draw_redraw);
  glutReshapeFunc (display_reshape);
  glutMouseFunc (interface_onMouse);
#if (MYFREEGLUT)
  glutMouseWheelFunc (interface_onMouse);
#endif
  glutKeyboardFunc (interface_onKeyDown);
  glutSpecialFunc (interface_onSKeyDown);
  glutSpecialUpFunc (interface_onKeyUp);
  glutMotionFunc (interface_motionActive);
  glutPassiveMotionFunc (interface_motionPassive);
  glutSetCursor (GLUT_CURSOR_NONE);
  

  drawworld_makeAll ();

  drawworld_createMinimap ();

  drawworld_activateArray ();

  drawworld_createTerrainLists ();

  glEnable (GL_DEPTH_TEST);
  glEnable (GL_LIGHTING);
  glEnable (GL_CULL_FACE);
  glShadeModel (GL_FLAT);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, black);
  /*
  glLightfv (GL_LIGHT1, GL_DIFFUSE, lightColor);
  glLightf (GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.1);
  glLightf (GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.01);
  glLightfv (GL_LIGHT0, GL_POSITION, lightPosition);
  */
  glEnable (GL_LIGHT0);
  if (l_quality == Q_HORRIBLE)
  {
    glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    glHint (GL_LINE_SMOOTH_HINT, GL_FASTEST);
    glHint (GL_POINT_SMOOTH_HINT, GL_FASTEST);
    glHint (GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
  }
  else
  {
    glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
    glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  }
  glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texColor);
  /*
  static float            fog_density = 0.001*2;
  glHint(GL_FOG_HINT, GL_NICEST);
  glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );
  glFogi(GL_FOG_MODE, GL_LINEAR);
  glFogi(GL_FOG_END, 2000);
  glFogf(GL_FOG_DENSITY, fog_density);
  glFogfv(GL_FOG_COLOR, fogColor);
  glEnable(GL_FOG);
  */
  
  glSelectBuffer (SELBUFFLEN, l_selbuff);
}

/* When moving, adjust the values that are only concerned by drawing */
void
draw_adjustVars (int elapsed, int action, Lobster *lobster)
{
  int i;
  float legs[4] = { 20., 20., 20., 20. }; /* step */
  float legd[4] = { -15., 5., 20., 45. }; /* last position before putting the leg a step forward */

  if (elapsed > lobster->steptime) /* end of move */
  {
    for (i = 0; i <= 3; i++)
      l_legp[i] = l_legp[i] - lobster->gone;
    lobster->angle = lobster->anglebase;
    lobster->gone = l_gonev = 0;
  }
  else if (elapsed == -1) /* motion init */
  {
    if (action == 0)  /* init of the game */
      lobster->angle = lobster->anglebase = 0.;
    if (action == 2)  /* turning left */
      lobster->anglebase = (int)(lobster->anglebase+90)%360;
    if (action == 3)  /* turning right */
      lobster->anglebase = (int)(lobster->anglebase-90)%360;
    /* printf("%f\n", lobster->anglebase); */
  }
  else /* moving */
  {
    lobster->gone = (float)elapsed / (float)lobster->steptime * lobster->stepsize;
    if (action == 4)
      lobster->gone = -(lobster->gone);
    for (i = 0; i <= 3; i++)
    {
      if (action != 4)
        if (lobster->legangle[i] >= legd[i])
          l_legp[i] += legs[i];
      if (action == 4)
        if (lobster->legangle[i] <= legd[i] - atan (legs[i] / LEGLONG) * raddeg)
          l_legp[i] -= legs[i];
      lobster->legangle[i] = l_legb[i] + atan ((lobster->gone - l_legp[i]) / LEGLONG) * raddeg;
    }
    lobster->phase = sin (elapsed * 2 * M_PI / lobster->steptime) * 20;
    if (action == 2)
      lobster->angle = -atan ((T_C - lobster->gone) / lobster->gone) * raddeg + lobster->anglebase;
    if (action == 3)
      lobster->angle = +atan ((T_C - lobster->gone) / lobster->gone) * raddeg + lobster->anglebase;

    if (lobster->steptime == LM_E)
      l_gonev = lobster->gone / 2.;
    else if (action == 2 || action == 3)
      l_gonev = lobster->gone + sin((float)elapsed / (float)lobster->steptime * M_PI)*20.;
    else
      l_gonev = lobster->gone;
  }
}

/* Set the view at a particuliar point */
void
draw_setView (int which)
{
  switch (which)
  {
    case 1:
      l_angle.x = 0.;
      l_angle.y = 0.;
      l_pose.x = 0;
      l_pose.z = 0;
      l_pose.y = -T_C * 1.5 + 1.;
      l_viewrot = 0.;
      break;
    case 2:
      l_angle.x = 0.;
      l_angle.y = 0.;
      l_pose.x = 0;
      l_pose.z = 0;
      l_pose.y = -T_C / 2.;
      l_viewrot = 0.;
      break;
    case 3:
      l_angle.x = -45.;
      l_angle.y = 0.;
      l_pose.x = 0;
      l_pose.z = 0;
      l_pose.y = -T_C / 2.;
      l_viewrot = 0.;
      break;
    case 4:
      l_angle.x = 45.;
      l_angle.y = 0.;
      l_pose.x = 0;
      l_pose.z = 0;
      l_pose.y = -T_C / 2.;
      l_viewrot = 0.;
      break;
    case 5:
      l_angle.x = 180.;
      l_angle.y = 0.;
      l_pose.x = 0;
      l_pose.z = 0;
      l_pose.y = T_C / 2.;
      l_viewrot = 0.;
      break;
    case 6:
      l_angle.x = 0.;
      l_angle.y = 0.;
      l_pose.x = 0.;
      l_pose.z = -40.;
      l_pose.y = T_C / 5.;
      l_viewrot = 0.;
      break;
    case 7:
      l_angle.x = 0.;
      l_angle.y = 0.;
      l_pose.x = 0.;
      l_pose.z = -45.;
      l_pose.y = -T_C / 5.;
      l_viewrot = M_PI / 3.;
      break;
  }
}

/* move the view horizontaly */
void
draw_moveViewH (float dx, float dy)
{
  float dx2, dy2;
  dx2 = sin (l_angle.x * degrad) * dy;
  dy2 = cos (l_angle.x * degrad) * dy;
  dx2 += sin (l_angle.x * degrad + M_PI / 2.) * dx;
  dy2 += cos (l_angle.x * degrad + M_PI / 2.) * dx;
  if (l_pose.x + dx2 < T_C / 2. - 1. && l_pose.x + dx2 > -T_C / 2. + 1. &&
      l_pose.y + dy2 < T_C / 2. - 1. && l_pose.y + dy2 > -T_C * 1.5 + 1.)
  {
    l_pose.x += dx2;
    l_pose.y += dy2;
  }
}

/* move the view verticaly */
void
draw_moveViewV (float dx, float dy)
{
  l_pose.x -= dx;
  if (l_pose.x >  T_C / 2. - 1.) l_pose.x =   T_C / 2. - 1.;
  if (l_pose.x < -T_C / 2. + 1.) l_pose.x =  -T_C / 2. + 1.;
  l_pose.z -= dy;
  if (l_pose.z > 0.) l_pose.z = 0.;
  if (l_pose.z < -50. + 3.) l_pose.z = -50. + 3.;
}

/* shade */
void
draw_shade (float dx, float dy)
{
  l_angle.x += dx;
  l_angle.y += dy;
}

/* become crazy (rotate the whole view) */
void
draw_rotateView (float sign)
{
  l_viewrot += 0.05 * sign;
}
